/*******************************************************************************
 Copyright 2011 Benjamin Fagin

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


    Read the included LICENSE.TXT for more information.
 ******************************************************************************/

package unquietcode.util;

/**
 * @author  Benjamin Fagin
 * @version 04-24-2011
 */
public abstract class SuperLoop implements SuperLoopFacade {
	private final Break BREAK = new Break();
	private final Continue CONTINUE = new Continue();
	private final String label;
	private boolean hasRun = false;

	public SuperLoop() {
		this(null);
	}

	public SuperLoop(String label) {
		this.label = label;
	}

	// Lifecycle Methods
	void head() { }
	boolean test() { return true; }
	abstract void during();
	void tail() { }
	void natural() { }
	void unnatural() { }

	// Runner
	public final void $() {
		hasRun = true;
		boolean natural = true;
		Break ignoreBreak = null;
		Continue ignoreContinue = null;

		try {
			// head, fires once before the loop tests or starts
			head();

			// loop test
			while(test()) {
				try {
					// loop body
					during();

				// on break
				} catch (Break b) {
					if (b == BREAK || b.label.equals(this.label)) {
						natural = false;
						break;
					} else {
						ignoreBreak = b;
						throw b;
					}

				// on continue
				} catch (Continue c) {
					if (c == CONTINUE || c.label.equals(this.label)) {
						continue;
					} else {
						ignoreContinue = c;
						throw c;
					}
				}
			}

			if (natural) {
				// loop completes naturally (test fails)
				natural();
			} else {
				// loop breaks out
				unnatural();
			}

			// tail, fires once at the very end ('finally')
			tail();

		} catch (Break b) {
			if (b == ignoreBreak) {
				throw b;
			} else {
				throw new LoopException("Invalid use of Break method outside of loop.");
			}
		} catch (Continue c) {
			if (c == ignoreContinue) {
				throw c;
			} else {
				throw new LoopException("Invalid use of Continue method outside of loop.");
			}
		}
	}

	public final boolean hasRun() {
		return hasRun;
	}


	// Escapes

	protected final void Break() {
		throw BREAK;
	}

	protected final void Break(String label) {
		if (label == null || label.equals(this.label)) {
			throw BREAK;
		} else {
			throw new Break(label);
		}
	}

	protected final void Continue() {
		throw CONTINUE;
	}

	protected final void Continue(String label) {
		if (label == null || label.equals(this.label)) {
			throw CONTINUE;
		} else {
			throw new Continue(label);
		}
	}


	private static class Break extends Command {
		Break() {
			super();
		}

		Break(String label) {
			super(label);
		}
	}
	private static class Continue extends Command {
		Continue() {
			super();
		}

		Continue(String label) {
			super(label);
		}
	}

	private static class Command extends LoopException {
		final String label;

		Command() {
			this(null);
		}

		Command(String label) {
			super(label == null
					? "Invalid label: null"
					: "Invalid label: '"+ label +"'");
			
			this.label = label;
		}
	}

	public static class LoopException extends RuntimeException {
		private LoopException(String message) {
			super(message);
		}
	}
}
