/*******************************************************************************
 Copyright 2011 Benjamin Fagin

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


    Read the included LICENSE.TXT for more information.
 ******************************************************************************/

package unquietcode.util;

import org.junit.Test;
import org.junit.internal.runners.statements.ExpectException;

import java.util.Random;

import static unquietcode.util.Shortcuts.out;

/**
 * @author  Benjamin Fagin
 * @version 04-24-2011
 */
public class SuperLoop_T {

	@Test
	public void smallestForm() {
		(new SuperLoop() {
			int lcv = 0;

			boolean test() {
				return lcv < 10;
			}

			void during() {
				// stuff

				lcv += 1;
			}
		}).$();
	}

	@Test
	public void facade() {
		SuperLoopFacade loop = new SuperLoop() {
			int lcv = 0;

			boolean test() {
				return (lcv < 10);
			}

			void during() {
				// stuff
				System.out.print("~run~");
				lcv += 1;
			}
		};

		SuperLoop reversed = (SuperLoop) loop;
		reversed.$();
	}


	@Test
	public void basic() {

		// no label
		(new SuperLoop() {
			int lcv;

			void head() {
				out("head");
				lcv = 0;
			}

			boolean test() {
				out("test");
				return (lcv < 10);
			}

			void during() {
				out("during (lcv == "+ lcv +")");
				lcv += 1;
			}

			void tail() {
				out("tail");
			}
		}).$();
	}

	@Test
	public void extendedLifecycle() {

		SuperLoopFacade loop = new SuperLoop() {
			Random gen = new Random();
			int lcv;

			void head() {
				out("head");
				lcv = 0;
			}

			boolean test() {
				out("test");
				return (lcv < 10);
			}

			void during() {
				out("during (lcv == "+ lcv +")");

				if (lcv == 9 && gen.nextBoolean()) {
					Break();
				}

				lcv += 1;
			}

			void natural() {
				out("natural ending");
			}

			void unnatural() {
				out("unnatural termination");
			}

			void tail() {
				out("tail");
			}
		};loop.$();
	}

	@Test
	public void nested() {
		// no label
		(new SuperLoop() {
			int lcv;

			void head() {
				out("head");
				lcv = 0;
			}

			boolean test() {
				out("test");
				return (lcv < 10);
			}

			void during() {
				out("during (lcv == "+ lcv +")");

				(new SuperLoop() {
					int lcv = 0;

					boolean test() {
						return (lcv < 3);
					}

					void during() {
						out("inner loop, count = " + lcv);
						lcv += 1;
					}
				}).$();

				lcv += 1;
			}

			void tail() {
				out("tail");
			}
		}).$();
	}

	@Test
	public void labeled() {
		(new SuperLoop("outer") {
			int lcv;

			void head() {
				out("head");
				lcv = 0;
			}

			boolean test() {
				out("test");
				return (lcv < 10);
			}

			void during() {
				out("during (lcv == "+ lcv +")");

				(new SuperLoop() {
					int lcv = 0;
					Random gen = new Random();

					boolean test() {
						return (lcv < 3);
					}

					void during() {
						out("inner loop, count = " + lcv);

						if (gen.nextInt(10) > 6) {
							Break("outer");
						} else if (lcv == 2 && gen.nextBoolean()) {
							if (gen.nextBoolean()) {
								Continue("outer");
							} else {
								Continue();
							}
						}

						lcv += 1;
					}

					void unnatural() {
						out("inner loop unnatural ending");
					}
				}).$();

				lcv += 1;
			}

			void tail() {
				out("tail");
			}

			void unnatural() {
				out("outer loop unnatural ending");
			}
		}).$();
	}



	@Test(expected = SuperLoop.LoopException.class)
	public void badLabel() {
		(new SuperLoop() {
			int lcv;

			void head() {
				out("head");
				lcv = 0;
			}

			boolean test() {
				out("test");
				return (lcv < 10);
			}

			void during() {
				out("during (lcv == "+ lcv +")");

				(new SuperLoop() {
					int lcv = 0;
					Random gen = new Random();

					boolean test() {
						return (lcv < 3);
					}

					void during() {
						out("inner loop, count = " + lcv);
						Break("outer");
					}

					void unnatural() {
						out("inner loop unnatural ending");
					}
				}).$();

				lcv += 1;
			}

			void tail() {
				out("tail");
			}

			void unnatural() {
				out("outer loop unnatural ending");
			}
		}).$();
	}

	@Test(expected = SuperLoop.LoopException.class)
	public void invalidEscape() {
		(new SuperLoop() {
			Random gen = new Random();

			boolean test() {
				out("test");
				return true;
			}

			void during() {
				out("during");

				if (gen.nextInt(10) > 4) {
					Break();
				}
			}

			void tail() {
				out("tail");    // won't print
			}

			void unnatural() {
				out("unnatural termination");
				Break();    // illegal
			}
		}).$();
	}
	

	// shortcut method for standard out
	private static void out(String string) {
		System.out.println(string);		
	}
}
